import gdb

'''try:
    import IPython
    has_shell = True
except:
    has_shell = False'''

from angrgdb import *

import sys
if sys.version_info >= (3, 0):
    long = int
    raw_input = input
else:
    long
    bytes = str


BANNER = " >>"

class AngrGDBError(RuntimeError):
    pass


def _to_int(x):
    try:
        return int(gdb.parse_and_eval(x).cast(gdb.lookup_type("long")))
    except BaseException as e:
        print (e)
        return None

def _prepare_args(args):
    return args.replace("\\", "\\\\").replace("\"", "\\\"")

def _prepare_shell(loc):
    sm = StateManager(StateShot(sync_brk=False))
    for k in _ctx.symbolics:
        if _ctx.symbolics[k] is None:
            sm.sim(k)
        else:
            sm.sim(k, _ctx.symbolics[k])
    loc["sm"] = sm
    loc["m"] = sm.simulation_manager()
    
    loc["find"] = _ctx.find
    loc["avoid"] = _ctx.avoid


class CommandsContext(object):
    def __init__(self):
        self.symbolics = {}
        self.find = []
        self.avoid = []


_ctx = CommandsContext()


class AngrGDBCommand(gdb.Command):
    '''
    Symbolic execution in GDB with angrdbg
    '''

    def __init__(self):
        super(AngrGDBCommand, self).__init__("angrgdb", gdb.COMMAND_USER, gdb.COMPLETE_NONE, True)


class AngrGDBShellCommand(gdb.Command):
    '''
    Open a python shell with a StateManager instance
    '''

    def __init__(self):
        super(AngrGDBShellCommand, self).__init__("angrgdb shell", gdb.COMMAND_USER)

    def invoke(self, arg, from_tty):
        global _ctx
        self.dont_repeat()
        
        if not from_tty:
            raise AngrGDBError("The angrgdb shell can be launched only from the tty")
                
        print (BANNER + " sm is a StateManager instance created from the current GDB state")
        print (BANNER + " m is a SimulationManager based on the sm state")
        print (BANNER + " find and avoid are the list of addresses")
        print (BANNER + " to find:", ", ".join(map(lambda x: "0x%x" % x, _ctx.find)))
        print (BANNER + " to avoid:", ", ".join(map(lambda x: "0x%x" % x, _ctx.avoid)))
                
        gdb.execute("py from angrgdb.commands import _prepare_shell; _prepare_shell(locals()); gdb.execute('pi')")


class AngrGDBResetCommand(gdb.Command):
    '''
    Reset the context fo angrgdb
    '''

    def __init__(self):
        super(
            AngrGDBResetCommand, self).__init__("angrgdb reset", gdb.COMMAND_USER)

    def invoke(self, arg, from_tty):
        global _ctx
        self.dont_repeat()

        _ctx = CommandsContext()


class AngrGDBSimCommand(gdb.Command):
    '''
    Set a memory/register as symbolic

    Usage: angrgdb sim <register name> [size]
           angrgdb sim <expression> [size]
    '''

    def __init__(self):
        super(
            AngrGDBSimCommand,
            self).__init__(
            "angrgdb sim",
            gdb.COMMAND_DATA)

    def _process_argv0(self, x):
        r = _to_int(x)
        if r:
            return r
        if x in load_project().arch.registers:
            return x
        raise AngrGDBError(
            "angrdbg sim: the first parameter is not an address or a register")

    def invoke(self, arg, from_tty):
        global _ctx
        self.dont_repeat()

        argv = gdb.string_to_argv(_prepare_args(arg))
        if len(argv) == 0:
            raise AngrGDBError("angrdbg sim: at least a parameter is needed")
        elif len(argv) == 1:
            _ctx.symbolics[self._process_argv0(argv[0])] = None
        else:
            siz = _to_int(argv[1])
            if siz is None:
                raise AngrGDBError(
                    "angrdbg sim: the second parameter (length) must be a number")
            _ctx.symbolics[self._process_argv0(argv[0])] = siz


class AngrGDBListCommand(gdb.Command):
    '''
    List all items that you setted as symbolic
    '''

    def __init__(self):
        super(
            AngrGDBListCommand,
            self).__init__(
            "angrgdb list",
            gdb.COMMAND_DATA)

    def invoke(self, arg, from_tty):
        global _ctx
        self.dont_repeat()

        for k in _ctx.symbolics:
            out = k
            if isinstance(k, int):
                out = "0x%x" % k
            if _ctx.symbolics[k] is not None:
                out += " " * (20 - len(out)) + "<%d>" % _ctx.symbolics[k]
            print (out)


class AngrGDBFindCommand(gdb.Command):
    '''
    Set the list of find targets

    Usage: angrgdb find <address0> <address1> ... <addressN>
    '''

    def __init__(self):
        super(
            AngrGDBFindCommand,
            self).__init__(
            "angrgdb find",
            gdb.COMMAND_DATA)

    def invoke(self, arg, from_tty):
        global _ctx
        self.dont_repeat()

        argv = gdb.string_to_argv(_prepare_args(arg))
        if len(argv) == 0:
            raise AngrGDBError("angrdbg find: at least a parameter is needed")

        _ctx.find = []
        for a in argv:
            addr = _to_int(a)
            if addr is None:
                raise AngrGDBError(
                    "angrdbg find: failed to convert '%s' to int" %
                    a)
            _ctx.find.append(addr)


class AngrGDBAvoidCommand(gdb.Command):
    '''
    Set the list of avoid targets

    Usage: angrgdb avoid <address0> <address1> ... <addressN>
    '''

    def __init__(self):
        super(
            AngrGDBAvoidCommand,
            self).__init__(
            "angrgdb avoid",
            gdb.COMMAND_DATA)

    def invoke(self, arg, from_tty):
        global _ctx
        self.dont_repeat()

        argv = gdb.string_to_argv(_prepare_args(arg))
        if len(argv) == 0:
            raise AngrGDBError("angrdbg avoid: at least a parameter is needed")

        _ctx.avoid = []
        for a in argv:
            addr = _to_int(a)
            if addr is None:
                raise AngrGDBError(
                    "angrdbg avoid: failed to convert '%s' to int" %
                    a)
            _ctx.avoid.append(addr)


class AngrGDBRunCommand(gdb.Command):
    '''
    Generate a state from the debugger state and run the exploration
    '''

    def __init__(self):
        super(
            AngrGDBRunCommand,
            self).__init__(
            "angrgdb run",
            gdb.COMMAND_DATA)

    def invoke(self, arg, from_tty):
        global _ctx
        self.dont_repeat()

        if len(_ctx.find) == 0:
            raise AngrGDBError("angrdbg run: the find list can't be empty")

        print (BANNER + " to find:", ", ".join(map(lambda x: "0x%x" % x, _ctx.find)))
        print (BANNER + " to avoid:", ", ".join(map(lambda x: "0x%x" % x, _ctx.avoid)))

        sm = StateManager(StateShot(sync_brk=False))
        for k in _ctx.symbolics:
            if _ctx.symbolics[k] is None:
                sm.sim(k)
            else:
                sm.sim(k, _ctx.symbolics[k])
        m = sm.simulation_manager()

        print (BANNER + " running the exploration...")
        m.explore(find=_ctx.find, avoid=_ctx.avoid)
        if len(m.found) == 0:
            raise AngrGDBError(
                "angrdbg run: valid state not found after exploration")

        conc = sm.concretize(m.found[0])
        print (BANNER + " results:\n")
        for k in _ctx.symbolics:
            out = k
            if isinstance(k, int):
                out = "0x%x" % k
            if _ctx.symbolics[k] is not None:
                out += " " * (20 - len(out)) + "<%d>" % _ctx.symbolics[k]
            print (out)
            out = conc[k]
            if isinstance(out, (int, long)):
                print ("   ==> 0x%x" % out)
            else:
                ro = repr(out)
                print ("   ==> %s" % ro[1:] if ro.startswith("b") else ro)
            print ()

        r = raw_input(BANNER + " do you want to write-back the results in GDB? [Y, n] ")
        r = r.strip().upper()
        if r == "Y" or r == "":
            print (BANNER + " syncing results with debugger...")
            sm.to_dbg(m.found[0])


AngrGDBCommand()
AngrGDBShellCommand()
AngrGDBResetCommand()
AngrGDBSimCommand()
AngrGDBListCommand()
AngrGDBFindCommand()
AngrGDBAvoidCommand()
AngrGDBRunCommand()



import sys
import struct
import re

try:
    import gdb
except ImportError:
    print("angrgdb: fatal error: not running inside GDB")
    exit(1)

if sys.version_info >= (3, 0):
    long = int
else:
    long
    bytes = str

from angrdbg import *


class GDBDebugger(Debugger):
    def __init__(self):
        self.inferior = gdb.selected_inferior()
        self.pid = 0
        self.base_addr = None
        self.efl_map = {}
        self.efl_map['CF'] = 1 << 0
        self.efl_map['PF'] = 1 << 2
        self.efl_map['AF'] = 1 << 4
        self.efl_map['ZF'] = 1 << 6
        self.efl_map['SF'] = 1 << 7
        self.efl_map['TF'] = 1 << 8
        self.efl_map['IF'] = 1 << 9
        self.efl_map['DF'] = 1 << 10
        self.efl_map['OF'] = 1 << 11

    def _get_vmmap(self):
        maps = []
        if self.pid == 0:
            pid = self.inferior.pid
        else:
            pid = self.pid
        
        mpath = "/proc/%s/maps" % pid
        # 00400000-0040b000 r-xp 00000000 08:02 538840  /path/to/file
        pattern = re.compile(
            "([0-9a-f]*)-([0-9a-f]*) ([rwxps-]*)(?: [^ ]*){3} *(.*)")

        out = open(mpath).read()

        matches = pattern.findall(out)
        if matches:
            for (start, end, perm, mapname) in matches:
                start = int(("0x%s" % start), 0)
                end = int(("0x%s" % end), 0)
                if mapname == "":
                    mapname = "mapped"
                mapperm = 0
                if "r" in perm:
                    mapperm |= SEG_PROT_R
                if "w" in perm:
                    mapperm |= SEG_PROT_W
                if "x" in perm:
                    mapperm |= SEG_PROT_X
                maps += [(start, end, mapperm, mapname)]
        return maps

    def _get_sections(self):
        base = self.image_base()
        info = gdb.execute("info file", to_string=True)
        # 0x0000000000000238 - 0x0000000000000254 is .interp
        pattern = re.compile("0x([0-9a-f]*) - 0x([0-9a-f]*) is (.*)")

        matches = pattern.findall(info)
        # don't get sections of shared libs
        matches = filter(lambda x: " in " not in x[2], matches)
        return map(lambda x: (int(x[0], 16), int(x[1], 16), x[2]), matches)

    # -------------------------------------
    def before_stateshot(self):
        self.vmmap = self._get_vmmap()
        self.base_addr = self.vmmap[0][0]
        sections = self._get_sections()

        for start, end, name in sections:
            if name == load_project().arch.got_section_name:
                self.got = (start, end)
            elif name == ".plt":
                self.plt = (start, end)
            elif name == ".idata":
                self.plt = (start, end)
        self.long_type = gdb.lookup_type("long")

    def after_stateshot(self, state):
        pass
    # -------------------------------------

    def is_active(self):
        return gdb.selected_thread() is not None

    # -------------------------------------
    def input_file(self):
        return open(gdb.current_progspace().filename, "rb")

    def image_base(self):
        if self.base_addr is None:
            self.before_stateshot()
        return self.base_addr

    # -------------------------------------
    def get_byte(self, addr):
        try:
            return int(self.inferior.read_memory(addr, 1).tobytes()[0])
        except BaseException:
            return None

    def get_word(self, addr):
        try:
            return struct.unpack(
                "<H", self.inferior.read_memory(addr, 2).tobytes())[0]
        except BaseException:
            return None

    def get_dword(self, addr):
        try:
            return struct.unpack(
                "<I", self.inferior.read_memory(addr, 4).tobytes())[0]
        except BaseException:
            return None

    def get_qword(self, addr):
        try:
            return struct.unpack("<Q", self.inferior.read_memory(addr, 8).tobytes())[0]
        except BaseException:
            return None

    def get_bytes(self, addr, size):
        try:
            return self.inferior.read_memory(addr, size).tobytes()
        except BaseException:
            return None

    def put_byte(self, addr, value):
        self.inferior.write_memory(addr, chr(value))

    def put_word(self, addr, value):
        self.inferior.write_memory(addr, struct.pack("<H", value))

    def put_dword(self, addr, value):
        self.inferior.write_memory(addr, struct.pack("<I", value))

    def put_qword(self, addr, value):
        self.inferior.write_memory(addr, struct.pack("<Q", value))

    def put_bytes(self, addr, value):
        self.inferior.write_memory(addr, value)

    # -------------------------------------
    def get_reg(self, name):
        if name == "efl" or name == "eflags":
            value = 0
            for f in self.efl_map:
                if f in str(gdb.parse_and_eval("$eflags")):
                    value |= self.efl_map[f]
            return value
        else:
            return int(gdb.parse_and_eval("$" + name).cast(self.long_type))

    def set_reg(self, name, value):
        if name == "efl":
            name = "eflags"
        gdb.execute("set $%s = %d" % (name, value))

    # -------------------------------------
    def step_into(self):
        gdb.execute("stepi", to_string=True)

    def run(self):
        gdb.execute("continue")

    def wait_ready(self):
        pass

    def refresh_memory(self):
        pass

    # -------------------------------------
    def seg_by_name(self, name):
        for start, end, perms, mname in self.vmmap:
            if name == mname:
                return Segment(name, start, end, perms)
        return None

    def seg_by_addr(self, addr):
        for start, end, perms, name in self.vmmap:
            if addr >= start and addr < end:
                return Segment(name, start, end, perms)
        return None

    def get_got(self):  # return tuple(start_addr, end_addr)
        return self.got

    def get_plt(self):  # return tuple(start_addr, end_addr)
        return self.plt
    
    def get_idata(self):  # return tuple(start_addr, end_addr)
        return self.idata

    # -------------------------------------
    def resolve_name(self, name):  # return None on fail
        try:
            res = gdb.execute("info address " + name, to_string=True)
            a = res.find(" is at 0x")
            b = res.find(" ", a + len(" is at 0x"))
            return int(res[a + len(" is at 0x"):b], 16)
        except BaseException:
            return None


register_debugger(GDBDebugger())



#!/usr/bin/env python

__author__ = "Andrea Fioraldi"
__copyright__ = "Copyright 2017, Andrea Fioraldi"
__license__ = "BSD 2-Clause"
__email__ = "andreafioraldi@gmail.com"

from setuptools import setup

VER = "1.0.11"

setup(
    name='angrgdb',
    version=VER,
    license=__license__,
    description='Use angr inside GDB. Create an angr state from the current debugger state. ',
    author=__author__,
    author_email=__email__,
    url='https://github.com/andreafioraldi/angrgdb',
    download_url = 'https://github.com/andreafioraldi/angrgdb/archive/' + VER + '.tar.gz',
    package_dir={'angrgdb': 'angrgdb'},
    packages=['angrgdb'],
    install_requires=['angrdbg'],
)
